@@ -39,10 +39,9 @@ namespace galera
         virtual ssize_t sendv(const WriteSetVector&, size_t,
                               gcs_act_type_t, bool, bool) = 0;
         virtual ssize_t send (const void*, size_t, gcs_act_type_t, bool) = 0;
-        virtual ssize_t replv(const WriteSetVector&, gcs_action& act, bool,
-                              const wsrep_seq_cb_t*)
-            = 0;
-        virtual ssize_t repl(gcs_action& act, bool) = 0;
+        virtual ssize_t replv(const WriteSetVector&,
+                              gcs_action& act, bool) = 0;
+        virtual ssize_t repl (gcs_action& act, bool) = 0;
         virtual void    caused(gu::GTID& gtid,
                                gu::datetime::Date& wait_until) = 0;
         virtual ssize_t schedule() = 0;
@@ -63,7 +62,18 @@ namespace galera
                              size_t          data_len) = 0;
         virtual void    get_stats(gcs_stats*) const = 0;
         virtual void    flush_stats() = 0;
-        virtual void    get_status(gu::Status&) const = 0;
+        virtual int     get_status(gu::Status&) const = 0;
+        virtual void    get_membership(wsrep_allocator_cb        alloc,
+                                       struct wsrep_membership** memb) const =0;
+        virtual int     fetch_pfs_info(wsrep_node_info_t** nodes,
+                                       uint32_t* size,
+                                       int32_t* my_index,
+                                       uint32_t max_version) = 0;
+        virtual int     fetch_pfs_stat(wsrep_node_stat_t** nodes,
+                                       uint32_t* size,
+                                       int32_t* my_index,
+                                       uint32_t max_version) = 0;
+
         /*! @throws NotFound */
         virtual void    param_set (const std::string& key,
                                    const std::string& value) = 0;
@@ -135,10 +145,10 @@ namespace galera
             return gcs_send(conn_, act, act_len, act_type, scheduled);
         }
 
-        ssize_t replv(const WriteSetVector& actv, struct gcs_action& act,
-                      bool scheduled, const wsrep_seq_cb_t* seq_cb)
+        ssize_t replv(const WriteSetVector& actv,
+                      struct gcs_action& act, bool scheduled)
         {
-            return gcs_replv(conn_, &actv[0], &act, scheduled, seq_cb);
+            return gcs_replv(conn_, &actv[0], &act, scheduled);
         }
 
         ssize_t repl(struct gcs_action& act, bool scheduled)
@@ -237,9 +247,33 @@ namespace galera
             return gcs_flush_stats(conn_);
         }
 
-        void get_status(gu::Status& status) const
+        int get_status(gu::Status& status) const
         {
-            gcs_get_status(conn_, status);
+            return gcs_get_status(conn_, status);
+        }
+
+        int fetch_pfs_info(wsrep_node_info_t** nodes,
+                           uint32_t* size,
+                           int32_t* my_index,
+                           uint32_t max_version)
+        {
+            return gcs_fetch_pfs_info(conn_, nodes, size, my_index,
+                                      max_version);
+        }
+
+        int fetch_pfs_stat(wsrep_node_stat_t** nodes,
+                           uint32_t* size,
+                           int32_t* my_index,
+                           uint32_t max_version)
+        {
+            return gcs_fetch_pfs_stat(conn_, nodes, size, my_index,
+                                      max_version);
+        }
+
+        void get_membership(wsrep_allocator_cb alloc,
+                            struct wsrep_membership** memb) const
+        {
+	    gcs_get_membership(conn_, alloc, memb);
         }
 
         void param_set (const std::string& key, const std::string& value)
@@ -305,23 +339,24 @@ namespace galera
         { return -ENOSYS; }
 
         ssize_t replv(const WriteSetVector& actv,
-                      gcs_action& act, bool scheduled, const wsrep_seq_cb_t*)
+                      gcs_action& act, bool scheduled)
         {
             ssize_t ret(set_seqnos(act));
 
             if (gu_likely(0 != gcache_ && ret > 0))
             {
                 assert (ret == act.size);
-                gu::byte_t* ptr(
-                    reinterpret_cast<gu::byte_t*>(gcache_->malloc(act.size)));
-                act.buf = ptr;
+                void* ptx;
+                act.buf = gcache_->malloc(act.size, ptx);
                 ssize_t copied(0);
                 for (int i(0); copied < act.size; ++i)
                 {
-                    memcpy (ptr + copied, actv[i].ptr, actv[i].size);
+                    ::memcpy(static_cast<gu::byte_t*>(ptx) + copied,
+                             actv[i].ptr, actv[i].size);
                     copied += actv[i].size;
                 }
                 assert (copied == act.size);
+                gcache_->drop_plaintext(act.buf);
             }
 
             return ret;
@@ -334,8 +369,10 @@ namespace galera
             if (gu_likely(0 != gcache_ && ret > 0))
             {
                 assert (ret == act.size);
-                void* const ptr(gcache_->malloc(act.size));
-                memcpy (ptr, act.buf, act.size);
+                void* ptx;
+                void* const ptr(gcache_->malloc(act.size, ptx));
+                memcpy (ptx, act.buf, act.size);
+                gcache_->drop_plaintext(ptr);
                 act.buf = ptr;
                 // no freeing here - initial act.buf belongs to the caller
             }
@@ -411,8 +448,32 @@ namespace galera
 
         void flush_stats() {}
 
-        void get_status(gu::Status& status) const
-        {}
+        int get_status(gu::Status& status) const
+        {
+	    return 0;
+	}
+
+        int fetch_pfs_info(wsrep_node_info_t** nodes,
+                           uint32_t* size,
+                           int32_t* my_index,
+                           uint32_t max_version)
+        {
+            return -ENOTCONN;
+        }
+
+        int fetch_pfs_stat(wsrep_node_stat_t** nodes,
+                           uint32_t* size,
+                           int32_t* my_index,
+                           uint32_t max_version)
+        {
+            return -ENOTCONN;
+        }
+
+        void get_membership(wsrep_allocator_cb        alloc,
+                            struct wsrep_membership** memb) const
+        {
+	    *memb = 0;
+        }
 
         void param_set (const std::string& key, const std::string& value)
         {}
