@@ -16,6 +16,7 @@
 #include "gcs_fifo_lite.hpp"
 #include "gcs_sm.hpp"
 #include "gcs_gcache.hpp"
+#include "gcs_error.hpp"
 
 #include <galerautils.h>
 #include <gu_logger.hpp>
@@ -30,6 +31,8 @@
 #include <errno.h>
 #include <assert.h>
 
+#include <cinttypes>
+
 const char* gcs_node_state_to_str (gcs_node_state_t state)
 {
     static const char* str[GCS_NODE_STATE_MAX + 1] =
@@ -116,6 +119,15 @@ __attribute__((__packed__));
 
 struct gcs_conn
 {
+    gcs_conn(gu::Config& conf,
+             gcache_t*   gcache,
+             gu::Progress<gcs_seqno_t>::Callback* progress_cb,
+             const char* node_name,
+             const char* inc_addr,
+             int repl_proto_ver,
+             int appl_proto_ver);
+    ~gcs_conn();
+
     gu::UUID group_uuid;
     char* my_name;
     char* channel;
@@ -126,7 +138,6 @@ struct gcs_conn
     gcs_conn_state_t  state;
 
     gu_config_t*      config;
-    bool              config_is_local;
     struct gcs_params params;
 
     gcache_t*    gcache;
@@ -245,65 +256,63 @@ struct gcs_repl_act
     { }
 };
 
-/*! Releases resources associated with parameters */
-static void
-_cleanup_params (gcs_conn_t* conn)
+gcs_conn::gcs_conn(gu::Config& conf,
+                   gcache_t*   cache,
+                   gu::Progress<gcs_seqno_t>::Callback* const progress_cb,
+                   const char* const node_name,
+                   const char* const inc_addr,
+                   int const repl_proto_ver,
+                   int const appl_proto_ver)
+    :
+    group_uuid(),
+    my_name(),
+    channel(),
+    socket(),
+    my_idx(),
+    memb_num(),
+    state(GCS_CONN_DESTROYED),
+    config(reinterpret_cast<gu_config_t*>(&conf)),
+    params(conf),
+    gcache(cache),
+    sm(),
+    local_act_id(),
+    global_seqno(),
+    repl_q(),
+    send_thread(),
+    recv_q(),
+    recv_q_size(),
+    recv_thread(),
+    timeout(),
+    fc_lock(),
+    stfc(),
+    stop_sent_(),
+    stop_count(),
+    queue_len(),
+    upper_limit(),
+    lower_limit(),
+    fc_offset(),
+    max_fc_state(),
+    stats_fc_stop_sent(),
+    stats_fc_cont_sent(),
+    stats_fc_received(),
+    conf_id(),
+    need_to_join(),
+    join_gtid(),
+    join_code(),
+    sync_sent_(),
+    core(),
+    vote_lock_(),
+    vote_cond_(),
+    vote_gtid_(),
+    vote_res_(),
+    vote_wait_(),
+    vote_err_(),
+    inner_close_count(),
+    outer_close_count(),
+    progress_cb_(progress_cb),
+    progress_()
 {
-    if (conn->config_is_local) gu_config_destroy(conn->config);
-}
-
-/*! Creates local configuration object if no external is submitted */
-static long
-_init_params (gcs_conn_t* conn, gu_config_t* conf)
-{
-    long rc;
-
-    conn->config = conf;
-    conn->config_is_local = false;
-
-    if (!conn->config) {
-        conn->config = gu_config_create();
-
-        if (conn->config) {
-            conn->config_is_local = true;
-        }
-        else {
-            rc = -ENOMEM;
-            goto enomem;
-        }
-    }
-
-    rc = gcs_params_init (&conn->params, conn->config);
-
-    if (!rc) return 0;
-
-    _cleanup_params (conn);
-
-enomem:
-
-    gu_error ("Parameter initialization failed: %s", strerror (-rc));
-
-    return rc;
-}
-
-/* Creates a group connection handle */
-gcs_conn_t*
-gcs_create (gu_config_t* const conf, gcache_t* const gcache,
-            gu::Progress<gcs_seqno_t>::Callback* const progress_cb,
-            const char* const node_name, const char* const inc_addr,
-            int const repl_proto_ver, int const appl_proto_ver)
-{
-    gcs_conn_t* conn = GU_CALLOC (1, gcs_conn_t);
-
-    if (!conn) {
-        gu_error ("Could not allocate GCS connection handle: %s",
-                  strerror (ENOMEM));
-        return NULL;
-    }
-
-    if (_init_params (conn, conf)) {
-        goto init_params_failed;
-    }
+    auto conn(this); // to minimize diff
 
     if (gcs_fc_init (&conn->stfc,
                      conn->params.recv_q_hard_limit,
@@ -314,7 +323,8 @@ gcs_create (gu_config_t* const conf, gcache_t* const gcache,
     }
 
     conn->state = GCS_CONN_DESTROYED;
-    conn->core  = gcs_core_create (conf, gcache, node_name, inc_addr,
+    conn->core  = gcs_core_create (conf, conn->gcache,
+                                   node_name, inc_addr,
                                    repl_proto_ver, appl_proto_ver);
     if (!conn->core) {
         gu_error ("Failed to create core.");
@@ -354,7 +364,6 @@ gcs_create (gu_config_t* const conf, gcache_t* const gcache,
     conn->global_seqno = 0;
     conn->fc_offset    = 0;
     conn->timeout      = GU_TIME_ETERNITY;
-    conn->gcache       = gcache;
     conn->max_fc_state = conn->params.sync_donor ?
         GCS_CONN_DONOR : GCS_CONN_JOINED;
 
@@ -366,7 +375,7 @@ gcs_create (gu_config_t* const conf, gcache_t* const gcache,
     conn->progress_cb_ = progress_cb;
     conn->progress_ = NULL;
 
-    return conn; // success
+    return; // success
 
 sm_create_failed:
 
@@ -383,14 +392,26 @@ repl_q_failed:
 core_create_failed:
 fc_init_failed:
 
-    _cleanup_params (conn);
+    gu_throw_fatal << "Failed to create GCS connection handle.";
+}
 
-init_params_failed:
-
-    gu_free (conn);
-
-    gu_error ("Failed to create GCS connection handle.");
-    return NULL; // failure
+/* Creates a group connection handle */
+gcs_conn_t*
+gcs_create (gu::Config& conf, gcache_t* gcache,
+            gu::Progress<gcs_seqno_t>::Callback* const progress_cb,
+            const char* const node_name, const char* const inc_addr,
+            int const repl_proto_ver, int const appl_proto_ver)
+{
+    try
+    {
+        return new gcs_conn(conf, gcache,
+                            progress_cb, node_name, inc_addr,
+                            repl_proto_ver, appl_proto_ver);
+    }
+    catch (...)
+    {
+        return nullptr;
+    }
 }
 
 long
@@ -426,7 +447,7 @@ gcs_check_error (int err, const char* warning)
     case -ENOTCONN:
     case -ECONNABORTED:
         if (NULL != warning) {
-            gu_warn ("%s: %d (%s)", warning, err, strerror(-err));
+            gu_info ("%s: %d (%s)", warning, err, gcs_error_str(-err));
         }
         err = 0;
         break;
@@ -456,7 +477,7 @@ gcs_fc_stop_begin (gcs_conn_t* conn)
                 !(err = gu_mutex_lock (&conn->fc_lock)));
 
     if (gu_unlikely(err)) {
-            gu_fatal ("Mutex lock failed: %d (%s)", err, strerror(err));
+            gu_fatal ("Mutex lock failed: %ld (%s)", err, strerror(err));
             abort();
     }
 
@@ -491,7 +512,8 @@ gcs_fc_stop_end (gcs_conn_t* conn)
             conn->stop_sent_dec(1);
         }
 
-        gu_debug ("SENDING FC_STOP (local seqno: %lld, fc_offset: %ld): %d",
+        gu_debug("SENDING FC_STOP (local seqno: %" PRId64
+                 ", fc_offset: %ld): %d",
                  conn->local_act_id, conn->fc_offset, ret);
     }
     else
@@ -521,7 +543,7 @@ gcs_fc_cont_begin (gcs_conn_t* conn)
                 !(err = gu_mutex_lock (&conn->fc_lock)));
 
     if (gu_unlikely(err)) {
-        gu_fatal ("Mutex lock failed: %d (%s)", err, strerror(err));
+        gu_fatal ("Mutex lock failed: %ld (%s)", err, strerror(err));
         abort();
     }
 
@@ -553,7 +575,8 @@ gcs_fc_cont_end (gcs_conn_t* conn)
             conn->stop_sent_inc(1);
         }
 
-        gu_debug ("SENDING FC_CONT (local seqno: %lld, fc_offset: %ld): %d",
+        gu_debug("SENDING FC_CONT (local seqno: %" PRId64
+                 ", fc_offset: %ld): %d",
                  conn->local_act_id, conn->fc_offset, ret);
     }
     else
@@ -661,16 +684,17 @@ gcs_shift_state (gcs_conn_t*      const conn,
 
     if (!allowed[new_state][old_state]) {
         if (old_state != new_state) {
-            gu_warn ("GCS: Shifting %s -> %s is not allowed (TO: %lld)",
-                     gcs_conn_state_str[old_state],
-                     gcs_conn_state_str[new_state], conn->global_seqno);
+            gu_warn("GCS: Shifting %s -> %s is not allowed (TO: %" PRId64 ")",
+                    gcs_conn_state_str[old_state],
+                    gcs_conn_state_str[new_state], conn->global_seqno);
         }
         return false;
     }
 
     if (old_state != new_state) {
-        gu_info ("Shifting %s -> %s (TO: %lld)", gcs_conn_state_str[old_state],
-                 gcs_conn_state_str[new_state], conn->global_seqno);
+        gu_info("Shifting %s -> %s (TO: %" PRId64 ")",
+                gcs_conn_state_str[old_state], gcs_conn_state_str[new_state],
+                conn->global_seqno);
         conn->state = new_state;
     }
 
@@ -740,8 +764,8 @@ gcs_become_primary (gcs_conn_t* conn)
     int ret;
 
     if ((ret = _release_flow_control (conn))) {
-        gu_fatal ("Failed to release flow control: %ld (%s)",
-                  ret, strerror(ret));
+        gu_fatal ("Failed to release flow control: %d (%s)",
+                  ret, gcs_error_str(ret));
         gcs_close (conn);
         abort();
     }
@@ -794,7 +818,7 @@ gcs_become_donor (gcs_conn_t* conn)
                         -EPROTO);
         if (err < 0 && !(err == -ENOTCONN || err == -EBADFD)) {
             gu_fatal ("Failed to send State Transfer Request rejection: "
-                      "%zd (%s)", err, (strerror (-err)));
+                      "%zd (%s)", err, (gcs_error_str (-err)));
             assert (0);
             return -ENOTRECOVERABLE; // failed to clear donor status,
         }
@@ -849,8 +873,8 @@ gcs_become_joined (gcs_conn_t* conn)
     if (GCS_CONN_JOINER == conn->state) {
         ret = _release_sst_flow_control (conn);
         if (ret < 0) {
-            gu_fatal ("Releasing SST flow control failed: %ld (%s)",
-                      ret, strerror (-ret));
+            gu_fatal ("Releasing SST flow control failed: %d (%s)",
+                      ret, gcs_error_str (-ret));
             abort();
         }
         conn->timeout = GU_TIME_ETERNITY;
@@ -865,7 +889,7 @@ gcs_become_joined (gcs_conn_t* conn)
         gu_debug("Become joined, FC offset %ld", conn->fc_offset);
         /* One of the cases when the node can become SYNCED */
         if ((ret = gcs_send_sync (conn))) {
-            gu_warn ("Sending SYNC failed: %ld (%s)", ret, strerror (-ret));
+            gu_warn ("Sending SYNC failed: %d (%s)", ret, gcs_error_str(-ret));
         }
     }
     else {
@@ -958,11 +982,12 @@ s_join (gcs_conn_t* conn)
         switch (err)
         {
         case -ENOTCONN:
-            gu_warn ("Sending JOIN failed: %d (%s). "
-                     "Will retry in new primary component.", err,strerror(-err));
+            gu_info("Sending JOIN failed: %s. "
+                    "Will retry in new primary component.",
+                    gcs_error_str(-err));
             return 0;
         default:
-            gu_error ("Sending JOIN failed: %d (%s).", err, strerror(-err));
+            gu_error("Sending JOIN failed: %d (%s).", err, gcs_error_str(-err));
             return err;
         }
     }
@@ -1107,7 +1132,7 @@ gcs_handle_act_conf (gcs_conn_t* conn, gcs_act_rcvd& rcvd)
     }
 
     if (old_state != conn->state) {
-        gu_info ("Restored state %s -> %s (%lld)",
+        gu_info ("Restored state %s -> %s (%" PRId64 ")",
                  gcs_conn_state_str[old_state], gcs_conn_state_str[conn->state],
                  conn->global_seqno);
     }
@@ -1116,7 +1141,7 @@ gcs_handle_act_conf (gcs_conn_t* conn, gcs_act_rcvd& rcvd)
     case GCS_CONN_JOINED:
         /* One of the cases when the node can become SYNCED */
         if ((ret = gcs_send_sync(conn)) < 0) {
-            gu_warn ("CC: sending SYNC failed: %ld (%s)", ret, strerror (-ret));
+            gu_warn ("CC: sending SYNC failed: %ld (%s)", ret, gcs_error_str (-ret));
         }
     break;
     case GCS_CONN_JOINER:
@@ -1138,7 +1163,7 @@ gcs_handle_act_state_req (gcs_conn_t*          conn,
 {
     if ((gcs_seqno_t)conn->my_idx == rcvd.id) {
         int const donor_idx = (int)rcvd.id; // to pacify valgrind
-        gu_debug("Got GCS_ACT_STATE_REQ to %i, my idx: %ld",
+        gu_debug("Got GCS_ACT_STATE_REQ to %i, my idx: %d",
                  donor_idx, conn->my_idx);
         // rewrite to pass global seqno for application
         rcvd.id = conn->global_seqno;
@@ -1157,7 +1182,7 @@ static long
 gcs_handle_state_change (gcs_conn_t*           conn,
                          const struct gcs_act* act)
 {
-    gu_debug ("Got '%s' dated %lld", gcs_act_type_to_str (act->type),
+    gu_debug ("Got '%s' dated %" PRId64, gcs_act_type_to_str (act->type),
               gcs_seqno_gtoh(*(gcs_seqno_t*)act->buf));
 
     void* buf = malloc (act->buf_len);
@@ -1389,7 +1414,7 @@ _close(gcs_conn_t* conn, bool join_recv_thread)
             /* if called from gcs_close(), we need to synchronize with
                gcs_recv_thread at this point */
             if ((ret = gu_thread_join (conn->recv_thread, NULL))) {
-                gu_error ("Failed to join recv_thread(): %d (%s)",
+                gu_error ("Failed to join recv_thread(): %ld (%s)",
                           -ret, strerror(-ret));
             }
             else {
@@ -1455,7 +1480,8 @@ static void *gcs_recv_thread (void *arg)
 
         if (gu_unlikely(ret <= 0)) {
 
-            gu_debug ("gcs_core_recv returned %d: %s", ret, strerror(-ret));
+            gu_debug("gcs_core_recv returned %zd: %s", ret,
+                     gcs_error_str(-ret));
 
             if (-ETIMEDOUT == ret && _handle_timeout(conn)) continue;
 
@@ -1502,7 +1528,7 @@ static void *gcs_recv_thread (void *arg)
             }
             if (gu_unlikely(ret < 0))
             { // error
-                gu_debug ("gcs_handle_actions returned %d: %s",
+                gu_debug ("gcs_handle_actions returned %zd: %s",
                           ret, strerror(-ret));
                 break;
             }
@@ -1576,8 +1602,8 @@ static void *gcs_recv_thread (void *arg)
                 }
 
                 if (gu_unlikely(send_stop) && (ret = gcs_fc_stop_end(conn))) {
-                    gu_error ("gcs_fc_stop() returned %d: %s",
-                              ret, strerror(-ret));
+                    gu_error ("gcs_fc_stop() returned %zd: %s",
+                              ret, gcs_error_str(-ret));
                     break;
                 }
             }
@@ -1605,7 +1631,7 @@ static void *gcs_recv_thread (void *arg)
         else if (conn->my_idx == rcvd.sender_idx)
         {
             gu_debug("Discarding: unordered local action not in repl_q: "
-                     "{ {%p, %zd, %s}, %d, %lld }.",
+                     "{ {%p, %zd, %s}, %d, %" PRId64 " }.",
                      rcvd.act.buf, rcvd.act.buf_len,
                      gcs_act_type_to_str(rcvd.act.type), rcvd.sender_idx,
                      rcvd.id);
@@ -1613,7 +1639,7 @@ static void *gcs_recv_thread (void *arg)
         else
         {
             gu_fatal ("Protocol violation: unordered remote action: "
-                      "{ {%p, %zd, %s}, %d, %lld }",
+                      "{ {%p, %zd, %s}, %d, % " PRId64 " }",
                       rcvd.act.buf, rcvd.act.buf_len,
                       gcs_act_type_to_str(rcvd.act.type), rcvd.sender_idx,
                       rcvd.id);
@@ -1632,7 +1658,7 @@ static void *gcs_recv_thread (void *arg)
         (void)_close(conn, false);
         gcs_shift_state (conn, GCS_CONN_CLOSED);
     }
-    gu_info ("RECV thread exiting %d: %s", ret, strerror(-ret));
+    gu_info ("RECV thread exiting %zd: %s", ret, strerror(-ret));
     return NULL;
 }
 
@@ -1649,7 +1675,7 @@ long gcs_open (gcs_conn_t* conn, const char* channel, const char* url,
 
     if ((ret = gcs_sm_enter (conn->sm, &tmp_cond, false, true)))
     {
-        gu_error("Failed to enter send monitor: %d (%s)", ret, strerror(-ret));
+        gu_error("Failed to enter send monitor: %ld (%s)", ret, strerror(-ret));
         return ret;
     }
 
@@ -1678,7 +1704,7 @@ long gcs_open (gcs_conn_t* conn, const char* channel, const char* url,
             gcs_core_close (conn->core);
         }
         else {
-            gu_error ("Failed to open channel '%s' at '%s': %d (%s)",
+            gu_error ("Failed to open channel '%s' at '%s': %ld (%s)",
                       channel, url, ret, strerror(-ret));
         }
     }
@@ -1712,7 +1738,7 @@ long gcs_close (gcs_conn_t *conn)
         /* _close() has already been called by gcs_recv_thread() and it
            is taking care of cleanup, just join the thread */
         if ((ret = gu_thread_join (conn->recv_thread, NULL))) {
-            gu_error ("Failed to join recv_thread(): %d (%s)",
+            gu_error ("Failed to join recv_thread(): %ld (%s)",
                       -ret, strerror(-ret));
         }
         else {
@@ -1729,11 +1755,10 @@ long gcs_close (gcs_conn_t *conn)
     return ret;
 }
 
-/* Frees resources associated with GCS connection handle */
-long gcs_destroy (gcs_conn_t *conn)
+gcs_conn::~gcs_conn()
 {
-    long err;
-
+    auto conn(this); // to minimize diff
+    int err;
     gu_cond_t tmp_cond;
     gu_cond_init (gu::get_cond_key(gu::GU_COND_KEY_GCS_DESTROY), &tmp_cond);
 
@@ -1747,7 +1772,7 @@ long gcs_destroy (gcs_conn_t *conn)
 
             gu_cond_destroy (&tmp_cond);
 
-            return -EBADFD;
+            gu_throw_error(EBADFD);
         }
 
         gcs_sm_leave (conn->sm);
@@ -1768,24 +1793,30 @@ long gcs_destroy (gcs_conn_t *conn)
 
     if ((err = gcs_fifo_lite_destroy (conn->repl_q))) {
         gu_debug ("Error destroying repl FIFO: %d (%s)", err, strerror(-err));
-        return err;
+        gu_throw_error(-err);
     }
 
     if ((err = gcs_core_destroy (conn->core))) {
         gu_debug ("Error destroying core: %d (%s)", err, strerror(-err));
-        return err;
+        gu_throw_error(-err);
     }
 
     gu_cond_destroy(&conn->vote_cond_);
     gu_mutex_destroy(&conn->vote_lock_);
     /* This must not last for long */
     while (gu_mutex_destroy (&conn->fc_lock));
+}
 
-    _cleanup_params (conn);
-
-    gu_free (conn);
-
-    return 0;
+/* Frees resources associated with GCS connection handle */
+long gcs_destroy (gcs_conn_t *conn)
+{
+    try {
+        delete conn;
+        return 0;
+    }
+    catch (...) {
+        return -1;
+    }
 }
 
 /* Puts action in the send queue and returns */
@@ -1906,9 +1937,10 @@ long gcs_replv (gcs_conn_t*          const conn,      //!<in
 
                 if (ret < 0) {
                     /* remove item from the queue, it will never be delivered */
-                    gu_warn ("Send action {%p, %zd, %s} returned %d (%s)",
-                             act->buf, act->size,gcs_act_type_to_str(act->type),
-                             ret, strerror(-ret));
+                    gu_debug(
+                        "Send action {%p, %" PRId32 ", %s} returned %ld (%s)",
+                        act->buf, act->size, gcs_act_type_to_str(act->type),
+                        ret, gcs_error_str(-ret));
 
                     if (!gcs_fifo_lite_remove (conn->repl_q)) {
                         gu_fatal ("Failed to remove unsent item from repl_q");
@@ -1958,7 +1990,7 @@ long gcs_replv (gcs_conn_t*          const conn,      //!<in
 
                     if (orig_buf != act->buf) // action was allocated in gcache
                     {
-                        gu_debug("Freeing gcache buffer %p after receiving %d",
+                        gu_debug("Freeing gcache buffer %p after receiving %ld",
                                  act->buf, ret);
                         gcs_gcache_free (conn->gcache, act->buf);
                         act->buf = orig_buf;
@@ -2144,19 +2176,19 @@ long gcs_recv (gcs_conn_t*        conn,
             if (conn->queue_len > 0) {
                 gu_warn ("Failed to send CONT message: %d (%s). "
                          "Attempts left: %ld",
-                         err, strerror(-err), conn->queue_len);
+                         err, gcs_error_str(-err), conn->queue_len);
             }
             else {
                 gu_fatal ("Last opportunity to send CONT message failed: "
                           "%d (%s). Aborting to avoid cluster lock-up...",
-                          err, strerror(-err));
+                          err, gcs_error_str(-err));
                 gcs_close(conn);
                 gu_abort();
             }
         }
         else if (gu_unlikely(send_sync) && (err = gcs_send_sync_end (conn))) {
             gu_warn ("Failed to send SYNC message: %d (%s). Will try later.",
-                     err, strerror(-err));
+                     err, gcs_error_str(-err));
         }
 
         return action->size;
@@ -2429,6 +2461,11 @@ gcs_get_stats (gcs_conn_t* conn, struct gcs_stats* stats)
     stats->fc_received = conn->stats_fc_received;
     stats->fc_active   = fc_active(conn);
     stats->fc_requested= conn->stop_sent_ > 0;
+
+    gcs_core_get_protocols(conn->core,
+                           stats->proto_appl,
+                           stats->proto_repl,
+                           stats->proto_gcs);
 }
 
 void
@@ -2701,9 +2738,10 @@ _set_max_throttle (gcs_conn_t* conn, const char* value)
     }
 }
 
-bool gcs_register_params (gu_config_t* const conf)
+void gcs_register_params (gu::Config& conf)
 {
-    return (gcs_params_register (conf) || gcs_core_register (conf));
+    gcs_params::register_params(conf);
+    gcs_core_register(conf);
 }
 
 long gcs_param_set  (gcs_conn_t* conn, const char* key, const char *value)
