@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2016 Codership Oy <info@codership.com>
+ * Copyright (C) 2008-2020 Codership Oy <info@codership.com>
  *
  * $Id$
  *
@@ -15,6 +15,7 @@
 #include "gcs_backend.hpp"
 #include "gcs_comp_msg.hpp"
 #include "gcs_code_msg.hpp"
+#include "gcs_error.hpp"
 #include "gcs_fifo_lite.hpp"
 #include "gcs_group.hpp"
 #include "gcs_gcache.hpp"
@@ -23,17 +24,23 @@
 #include <gu_logger.hpp>
 #include <gu_serialize.hpp>
 #include <gu_debug_sync.hpp>
+#include <gu_thread_keys.hpp>
 
 #include <string.h> // for mempcpy
 #include <errno.h>
 
+#include <cinttypes>
+
 using namespace gcs::core;
 
-bool
-gcs_core_register (gu_config_t* conf)
+void
+gcs_core_register(gu::Config& conf)
 {
-    gcs_group_register(reinterpret_cast<gu::Config*>(conf));
-    return (gcs_backend_register(conf));
+    gcs_group::register_params(conf);
+    if (gcs_backend_register(reinterpret_cast<gu_config_t*>(&conf)))
+    {
+        gu_throw_fatal << "Could not register backend parmeters";
+    }
 }
 
 const size_t CORE_FIFO_LEN = (1 << 10); // 1024 elements (no need to have more)
@@ -51,9 +58,21 @@ core_state_t;
 
 struct gcs_core
 {
+    gcs_core(gu::Config&  conf,
+             gcache_t*    cache,
+             const char*  node_name,
+             const char*  inc_addr,
+             int          repl_proto_ver,
+             int          appl_proto_ver,
+             int          gcs_proto_ver = GCS_PROTO_MAX);
+    ~gcs_core();
+
     gu_config_t*    config;
     gcache_t*       cache;
 
+    /* group context */
+    gcs_group_t     group;
+
     /* connection per se */
     long            prim_comp_no;
     core_state_t    state;
@@ -78,9 +97,6 @@ struct gcs_core
     /* local action FIFO */
     gcs_fifo_lite_t* fifo;
 
-    /* group context */
-    gcs_group_t     group;
-
     /* backend part */
     size_t          msg_size;
     gcs_backend_t   backend;   // message IO context
@@ -109,24 +125,37 @@ typedef struct causal_act
     gu_cond_t*   cond;
 } causal_act_t;
 
-gcs_core_t*
-gcs_core_create (gu_config_t* const conf,
-                 gcache_t*    const cache,
-                 const char*  const node_name,
-                 const char*  const inc_addr,
-                 int          const repl_proto_ver,
-                 int          const appl_proto_ver,
-                 int          const gcs_proto_ver)
+gcs_core::gcs_core(gu::Config&  conf,
+                   gcache_t*    cache,
+                   const char*  node_name,
+                   const char*  inc_addr,
+                   int          repl_proto_ver,
+                   int          appl_proto_ver,
+                   int          gcs_proto_ver)
+    :
+    config(reinterpret_cast<gu_config_t*>(&conf)),
+    cache(cache),
+    group(conf, cache, node_name, inc_addr,
+          gcs_proto_ver, repl_proto_ver,appl_proto_ver),
+    prim_comp_no(),
+    state(),
+    proto_ver(),
+    send_lock(),
+    send_buf(),
+    send_buf_len(),
+    send_act_no(),
+    recv_msg(),
+    code_msg_buf(),
+    fifo(),
+    msg_size(),
+    backend()
+#ifdef GCS_CORE_TESTING
+    ,ls()        // to lock-step in unit tests
+    ,state_uuid()
+#endif
 {
-    assert (conf);
-
-    gcs_core_t* core = GU_CALLOC (1, gcs_core_t);
-
-    if (NULL != core) {
-
-        core->config = conf;
-        core->cache  = cache;
-
+    auto core(this); // to minimize diff
+    {
         // Need to allocate something, otherwise Spread 3.17.3 freaks out.
         core->recv_msg.buf = gu_malloc(CORE_INIT_BUF_SIZE);
         if (core->recv_msg.buf) {
@@ -141,23 +170,18 @@ gcs_core_create (gu_config_t* const conf,
                 core->fifo = gcs_fifo_lite_create (CORE_FIFO_LEN,
                                                    sizeof (core_act_t));
                 if (core->fifo) {
-                    gu_mutex_init  (&core->send_lock, NULL);
+                    gu_mutex_init(
+                        gu::get_mutex_key(gu::GU_MUTEX_KEY_GCS_CORE_SEND),
+                        &core->send_lock);
                     core->proto_ver = -1;
                     // ^^^ shall be bumped in gcs_group_act_conf()
-
-                    gcs_group_init (&core->group,
-                                    reinterpret_cast<gu::Config*>(conf), cache,
-                                    node_name, inc_addr,
-                                    gcs_proto_ver, repl_proto_ver,appl_proto_ver
-                        );
-
                     core->state = CORE_CLOSED;
                     core->send_act_no = 1; // 0 == no actions sent
 #ifdef GCS_CORE_TESTING
                     gu_lock_step_init (&core->ls);
                     core->state_uuid = GU_UUID_NIL;
 #endif
-                    return core; // success
+                    return; // success
                 }
 
                 gu_free (core->send_buf);
@@ -165,11 +189,27 @@ gcs_core_create (gu_config_t* const conf,
 
             gu_free (core->recv_msg.buf);
         }
-
-        gu_free (core);
     }
 
-    return NULL; // failure
+    gu_throw_fatal << "Failed to initialize GCS core";
+}
+
+gcs_core_t*
+gcs_core_create (gu::Config&  conf,
+                 gcache_t*    const cache,
+                 const char*  const node_name,
+                 const char*  const inc_addr,
+                 int          const repl_proto_ver,
+                 int          const appl_proto_ver,
+                 int          const gcs_proto_ver)
+{
+    try {
+        return new gcs_core(conf, cache, node_name, inc_addr,
+                            repl_proto_ver, appl_proto_ver, gcs_proto_ver);
+    }
+    catch (...) {
+        return nullptr;
+    }
 }
 
 long
@@ -216,14 +256,14 @@ gcs_core_open (gcs_core_t* core,
             core->state = CORE_NON_PRIMARY;
         }
         else {
-            gu_error ("Failed to open backend connection: %d (%s)",
+            gu_error ("Failed to open backend connection: %ld (%s)",
                       ret, strerror(-ret));
             core->backend.destroy (&core->backend);
         }
 
     }
     else {
-        gu_error ("Failed to initialize backend using '%s': %d (%s)",
+        gu_error ("Failed to initialize backend using '%s': %ld (%s)",
                   url, ret, strerror(-ret));
     }
 
@@ -328,6 +368,8 @@ gcs_core_send (gcs_core_t*          const conn,
 
     assert (action != NULL);
     assert (act_size > 0);
+    /* actions passing this way will be put in cache */
+    assert (gcs_act_in_cache(act_type));
 
     /*
      * Action header will be replicated with every message.
@@ -351,7 +393,7 @@ gcs_core_send (gcs_core_t*          const conn,
     }
     else {
         ret = core_error (conn->state);
-        gu_error ("Failed to access core FIFO: %d (%s)", ret, strerror (-ret));
+        gu_error ("Failed to access core FIFO: %zd (%s)", ret, strerror (-ret));
         return ret;
     }
 
@@ -486,7 +528,7 @@ core_msg_recv (gcs_backend_t* backend, gcs_recv_msg_t* recv_msg,
         /* sometimes - like in case of component message, we may need to
          * do reallocation 2 times. This should be fixed in backend */
         void* msg = gu_realloc (recv_msg->buf, ret);
-        gu_debug ("Reallocating buffer from %d to %d bytes",
+        gu_debug ("Reallocating buffer from %d to %ld bytes",
                   recv_msg->buf_len, ret);
         if (msg) {
             /* try again */
@@ -500,7 +542,7 @@ core_msg_recv (gcs_backend_t* backend, gcs_recv_msg_t* recv_msg,
         }
         else {
             /* realloc unsuccessfull, old recv_buf remains */
-            gu_error ("Failed to reallocate buffer to %d bytes", ret);
+            gu_error ("Failed to reallocate buffer to %ld bytes", ret);
             ret = -ENOMEM;
             break;
         }
@@ -509,7 +551,7 @@ core_msg_recv (gcs_backend_t* backend, gcs_recv_msg_t* recv_msg,
     assert(recv_msg->buf);
 
     if (gu_unlikely(ret < 0)) {
-        gu_debug ("returning %d: %s\n", ret, strerror(-ret));
+        gu_debug ("returning %ld: %s\n", ret, strerror(-ret));
     }
 
     return ret;
@@ -570,7 +612,7 @@ core_handle_act_msg (gcs_core_t*          core,
 #else
             assert (NULL == act->act.buf);
 #endif
-            act->sender_idx = msg->sender_idx;
+            assert(act->sender_idx == msg->sender_idx);
 
             if (gu_likely(!my_msg)) {
                 /* foreign action, must be passed from gcs_group */
@@ -593,8 +635,10 @@ core_handle_act_msg (gcs_core_t*          core,
                     /* NOTE! local_act cannot be used after this point */
                     /* sanity check */
                     if (gu_unlikely(sent_act_id != frg.act_id)) {
-                        gu_fatal ("FIFO violation: expected sent_act_id %lld "
-                                  "found %lld", sent_act_id, frg.act_id);
+                        gu_fatal("FIFO violation: expected sent_act_id %" PRId64
+                                 " "
+                                 "found %" PRId64,
+                                 sent_act_id, frg.act_id);
                         ret = -ENOTRECOVERABLE;
                     }
                     if (gu_unlikely(act->act.buf_len != ret)) {
@@ -637,7 +681,7 @@ core_handle_act_msg (gcs_core_t*          core,
                 ret = gcs_group_handle_state_request (group, act);
                 assert (ret <= 0 || ret == act->act.buf_len);
 #ifdef GCS_FOR_GARB
-                if (ret < 0) gu_fatal ("Handling state request failed: %d",ret);
+                if (ret < 0) gu_fatal ("Handling state request failed: %ld",ret);
                 act->act.buf = NULL;
             }
             else {
@@ -661,7 +705,7 @@ core_handle_act_msg (gcs_core_t*          core,
     }
     else {
         /* Non-primary conf, foreign message - ignore */
-        gu_warn ("Action message in non-primary configuration from "
+        gu_info ("Action message in non-primary configuration from "
                  "member %d", msg->sender_idx);
         ret = 0;
     }
@@ -764,6 +808,17 @@ core_handle_vote_msg (gcs_core_t*          core,
     return 0;
 }
 
+/*! Common things to do on detected inconsistency */
+static int
+core_handle_inconsistency(gcs_core_t* core, struct gcs_act* act)
+{
+    core->state  = CORE_NON_PRIMARY;
+    act->buf     = NULL;
+    act->buf_len = 0;
+    act->type    = GCS_ACT_INCONSISTENCY;
+    return -ENOTRECOVERABLE;
+}
+
 /*!
  * Helper for gcs_core_recv(). Handles GCS_MSG_COMPONENT.
  *
@@ -781,7 +836,7 @@ core_handle_comp_msg (gcs_core_t*          const core,
     assert (GCS_MSG_COMPONENT == msg->type);
 
     if (msg->size < (ssize_t)sizeof(gcs_comp_msg_t)) {
-        gu_error ("Malformed component message (size %zd < %zd). Ignoring",
+        gu_error ("Malformed component message (size %d < %zu). Ignoring",
                   msg->size, sizeof(gcs_comp_msg_t));
         return 0;
     }
@@ -800,7 +855,7 @@ core_handle_comp_msg (gcs_core_t*          const core,
 
         ret = gcs_group_act_conf (group, rcvd, &core->proto_ver);
         if (ret < 0) {
-            gu_fatal ("Failed create PRIM CONF action: %d (%s)",
+            gu_fatal ("Failed create PRIM CONF action: %zd (%s)",
                       ret, strerror (-ret));
             assert (0);
             ret = -ENOTRECOVERABLE;
@@ -823,11 +878,25 @@ core_handle_comp_msg (gcs_core_t*          const core,
                                           &uuid,
                                           sizeof(uuid),
                                           GCS_MSG_STATE_UUID);
-                if (ret < 0) {
+                if (ret < 0)
+                {
                     // if send() failed, it means new configuration change
                     // is on the way. Probably should ignore.
-                    gu_warn ("Failed to send state UUID: %d (%s)",
-                             ret, strerror (-ret));
+                    switch (-ret)
+                    {
+                    case EAGAIN:
+                        gu_info("Temporary failure in sending state UUID, "
+                                "will try again in next primary component");
+                        break;
+                    case ENOTCONN:
+                        gu_info("Failed to send state UUID: Connection to "
+                                "cluster was closed");
+                        break;
+                    default:
+                        gu_warn("Failed to send state UUID: %zd (%s)", ret,
+                                gcs_error_str(-ret));
+                        break;
+                    }
                 }
                 else {
                     gu_info ("STATE_EXCHANGE: sent state UUID: "
@@ -852,7 +921,7 @@ core_handle_comp_msg (gcs_core_t*          const core,
                     assert(act->buf == NULL);
                     assert(act->buf_len == 0);
                     act->type = GCS_ACT_ERROR;
-                    gu_debug("comp msg error in core %d", -ret);
+                    gu_debug("comp msg error in core %ld", -ret);
                 }
             }
             else {                               // regular non-prim
@@ -862,7 +931,7 @@ core_handle_comp_msg (gcs_core_t*          const core,
             if (GCS_GROUP_NON_PRIMARY == ret) { // no error in comp msg
                 ret = gcs_group_act_conf (group, rcvd, &core->proto_ver);
                 if (ret < 0) {
-                    gu_fatal ("Failed create NON-PRIM CONF action: %d (%s)",
+                    gu_fatal ("Failed create NON-PRIM CONF action: %ld (%s)",
                               ret, strerror (-ret));
                     assert (0);
                     ret = -ENOTRECOVERABLE;
@@ -874,6 +943,9 @@ core_handle_comp_msg (gcs_core_t*          const core,
         }
         assert (ret == act->buf_len || ret < 0);
         break;
+    case GCS_GROUP_INCONSISTENT:
+        ret = core_handle_inconsistency(core, act);
+        break;
     case GCS_GROUP_WAIT_STATE_MSG:
         gu_fatal ("Internal error: gcs_group_handle_comp() returned "
                   "WAIT_STATE_MSG. Can't continue.");
@@ -881,7 +953,7 @@ core_handle_comp_msg (gcs_core_t*          const core,
         assert(0);
         // fall through
     default:
-        gu_fatal ("Failed to handle component message: %d (%s)!",
+        gu_fatal ("Failed to handle component message: %ld (%s)!",
                   ret, strerror (-ret));
         assert(0);
     }
@@ -932,7 +1004,7 @@ core_handle_uuid_msg (gcs_core_t*     core,
                         // This may happen if new configuraiton chage goes on.
                         // What shall we do in this case? Is it unrecoverable?
                         gu_error ("STATE EXCHANGE: failed for: " GU_UUID_FORMAT
-                                 ": %d (%s)",
+                                 ": %zd (%s)",
                                  GU_UUID_ARGS(state_uuid), ret, strerror(-ret));
                     }
                     gcs_state_msg_destroy (state);
@@ -948,7 +1020,7 @@ core_handle_uuid_msg (gcs_core_t*     core,
             break;
         default:
             assert(ret < 0);
-            gu_error ("Failed to handle state UUID: %d (%s)",
+            gu_error ("Failed to handle state UUID: %zd (%s)",
                       ret, strerror (-ret));
         }
     }
@@ -974,7 +1046,9 @@ core_handle_state_msg (gcs_core_t*          core,
     if (GCS_GROUP_WAIT_STATE_MSG == gcs_group_state (group))
     {
         if (gu_mutex_lock (&core->send_lock)) abort();
-        ret = gcs_group_handle_state_msg (group, msg);
+        // cast to int is needed to distinguish between positive enums and
+        // negative error codes
+        ret = int(gcs_group_handle_state_msg (group, msg));
 
         switch (ret) {
         case GCS_GROUP_PRIMARY:
@@ -997,7 +1071,7 @@ core_handle_state_msg (gcs_core_t*          core,
 
             ret = gcs_group_act_conf (group, rcvd, &core->proto_ver);
             if (ret < 0) {
-                gu_fatal ("Failed create CONF action: %d (%s)",
+                gu_fatal ("Failed create CONF action: %zd (%s)",
                           ret, strerror (-ret));
                 assert (0);
                 ret = -ENOTRECOVERABLE;
@@ -1008,9 +1082,12 @@ core_handle_state_msg (gcs_core_t*          core,
             // waiting for more state messages
             ret = 0;
             break;
+        case GCS_GROUP_INCONSISTENT:
+            ret = core_handle_inconsistency(core, &rcvd->act);
+            break;
         default:
             assert (ret < 0);
-            gu_error ("Failed to handle state message: %d (%s)",
+            gu_error ("Failed to handle state message: %zd (%s)",
                       ret, strerror (-ret));
         }
         gu_mutex_unlock (&core->send_lock);
@@ -1098,7 +1175,7 @@ core_msg_to_action (gcs_core_t*          core,
             }
             break;
         default:
-            gu_error ("Iternal error. Unexpected message type %s from %ld",
+            gu_error ("Iternal error. Unexpected message type %s from %d",
                       gcs_msg_type_string[msg->type], msg->sender_idx);
             assert (0);
             ret = -EPROTO;
@@ -1112,7 +1189,10 @@ core_msg_to_action (gcs_core_t*          core,
         }
     }
     else {
-        gu_warn ("%s message from member %ld in non-primary configuration. "
+        /* Messages which were sent just before cluster partitioning may
+         * be delivered in the following non-primary configuration. This
+         * is expected behavior, so info log level is enough. */
+        gu_info ("%s message from member %d in non-primary configuration. "
                  "Ignored.", gcs_msg_type_string[msg->type], msg->sender_idx);
     }
 
@@ -1124,7 +1204,7 @@ static long core_msg_causal(gcs_core_t* conn,
 {
     if (gu_unlikely(msg->size != sizeof(causal_act_t)))
     {
-        gu_error("invalid causal act len %ld, expected %ld",
+        gu_error("invalid causal act len %d, expected %zu",
                  msg->size, sizeof(causal_act_t));
         return -EPROTO;
     }
@@ -1202,17 +1282,17 @@ ssize_t gcs_core_recv (gcs_core_t*          conn,
         case GCS_MSG_COMPONENT:
             ret = core_handle_comp_msg (conn, recv_msg, recv_act);
             // assert (ret >= 0); // hang on error in debug mode
-            assert (ret == recv_act->act.buf_len || ret <= 0);
+            assert (ret == recv_act->act.buf_len || ret < 0);
             break;
         case GCS_MSG_STATE_UUID:
             ret = core_handle_uuid_msg (conn, recv_msg);
             // assert (ret >= 0); // hang on error in debug mode
-            ret = 0;           // continue waiting for state messages
+            ret = 0;              // continue waiting for state messages
             break;
         case GCS_MSG_STATE_MSG:
             ret = core_handle_state_msg (conn, recv_msg, recv_act);
-            assert (ret >= 0); // hang on error in debug mode
-            assert (ret == recv_act->act.buf_len);
+            // assert (ret >= 0); // hang on error in debug mode
+            assert (ret == recv_act->act.buf_len || ret < 0);
             break;
         case GCS_MSG_JOIN:
         case GCS_MSG_SYNC:
@@ -1265,7 +1345,12 @@ out:
 
         if (-ENOTRECOVERABLE == ret) {
             conn->backend.close(&conn->backend);
-            gu_abort();
+            if (GCS_ACT_INCONSISTENCY != recv_act->act.type) {
+                /* inconsistency event must be passed up */
+                gu_fatal("Unrecoverable error happened above. Aborting...");
+                usleep(1000000); // give it a second
+                gu_abort();
+            }
         }
     }
 
@@ -1291,12 +1376,11 @@ long gcs_core_close (gcs_core_t* core)
     return ret;
 }
 
-long gcs_core_destroy (gcs_core_t* core)
+static int
+core_destroy(gcs_core_t* core)
 {
     core_act_t* tmp;
 
-    if (!core) return -EBADFD;
-
     if (gu_mutex_lock (&core->send_lock)) return -EBADFD;
     {
         if (CORE_CLOSED != core->state) {
@@ -1324,7 +1408,6 @@ long gcs_core_destroy (gcs_core_t* core)
         gcs_fifo_lite_pop_head (core->fifo);
     }
     gcs_fifo_lite_destroy (core->fifo);
-    gcs_group_free (&core->group);
 
     /* free buffers */
     gu_free (core->recv_msg.buf);
@@ -1334,11 +1417,28 @@ long gcs_core_destroy (gcs_core_t* core)
     gu_lock_step_destroy (&core->ls);
 #endif
 
-    gu_free (core);
-
     return 0;
 }
 
+gcs_core::~gcs_core()
+{
+    int const ret(core_destroy(this));
+    if (ret) {
+        gu_throw_error(ret) << "GCS core destructor failed";
+    }
+}
+
+long gcs_core_destroy (gcs_core_t* core)
+{
+    try {
+        delete core;
+        return 0;
+    }
+    catch (...) {
+        return -1;
+    }
+}
+
 int
 gcs_core_proto_ver (const gcs_core_t* conn)
 {
@@ -1503,8 +1603,8 @@ gcs_core_caused (gcs_core_t* core, gu::GTID& gtid)
     gu_cond_t    cond;
     causal_act_t act = {&act_id, &act_uuid, &error, &mtx, &cond};
 
-    gu_mutex_init (&mtx, NULL);
-    gu_cond_init  (&cond, NULL);
+    gu_mutex_init (gu::get_mutex_key(gu::GU_MUTEX_KEY_GCS_CORE_CAUSED), &mtx);
+    gu_cond_init  (gu::get_cond_key(gu::GU_COND_KEY_GCS_CORE_CAUSED), &cond);
     gu_mutex_lock (&mtx);
     {
         long ret = core_msg_send_retry (core,
@@ -1557,16 +1657,72 @@ gcs_core_param_get (gcs_core_t* core, const char* key)
     }
 }
 
-void gcs_core_get_status(gcs_core_t* core, gu::Status& status)
+int gcs_core_get_status(gcs_core_t* core, gu::Status& status)
 {
+    int rc = 0;
     if (gu_mutex_lock(&core->send_lock))
-        gu_throw_fatal << "could not lock mutex";
+    {
+        return -ENOTRECOVERABLE;
+    }
     if (core->state < CORE_CLOSED)
     {
         gcs_group_get_status(&core->group, status);
         core->backend.status_get(&core->backend, status);
     }
     gu_mutex_unlock(&core->send_lock);
+    return rc;
+}
+
+int gcs_core_fetch_pfs_info(gcs_core_t*         core,
+                            wsrep_node_info_t** nodes,
+                            uint32_t*           size,
+                            int32_t*            my_index,
+                            uint32_t            max_version)
+{
+    int rc = -ENOTCONN;
+    if (gu_mutex_lock(&core->send_lock))
+    {
+        return -ENOTRECOVERABLE;
+    }
+    if (core->state < CORE_CLOSED)
+    {
+        rc = gcs_group_fetch_pfs_info(&core->group, nodes, size, my_index,
+                                      max_version);
+    }
+    gu_mutex_unlock(&core->send_lock);
+    return rc;
+}
+
+int gcs_core_fetch_pfs_stat(gcs_core_t*         core,
+                            wsrep_node_stat_t** nodes,
+                            uint32_t*           size,
+                            int32_t*            my_index,
+                            uint32_t            max_version)
+{
+    int rc = -ENOTCONN;
+    if (gu_mutex_lock(&core->send_lock))
+    {
+        return -ENOTRECOVERABLE;
+    }
+    if (core->state < CORE_CLOSED)
+    {
+        rc = gcs_group_fetch_pfs_stat(&core->group, nodes, size, my_index,
+                                      max_version);
+    }
+    gu_mutex_unlock(&core->send_lock);
+    return rc;
+}
+
+void gcs_core_get_membership(const gcs_core_t* const   core,
+                             wsrep_allocator_cb const  alloc,
+                             struct wsrep_membership** memb)
+{
+    gcs_group_get_membership(core->group, alloc, memb);
+}
+
+void gcs_core_get_protocols(gcs_core_t* core, int& appl, int& repl, int& gcs)
+{
+    core->group.get_protocols(appl, repl, gcs);
 }
 
 #ifdef GCS_CORE_TESTING
@@ -1607,4 +1763,10 @@ gcs_core_get_fifo (gcs_core_t* core)
     return core->fifo;
 }
 
+gcache_t*
+gcs_core_get_gcache(gcs_core_t* core)
+{
+    return core->cache;
+}
+
 #endif /* GCS_CORE_TESTING */
